@@ -1,53 +1,159 @@
 import asyncio
-import collections
+import codecs
 import io
-import json
-import ssl
+import re
 import sys
 import traceback
 import warnings
-from collections import namedtuple
 from hashlib import md5, sha1, sha256
-from http.cookies import CookieError, Morsel
+from http.cookies import CookieError, Morsel, SimpleCookie
+from types import MappingProxyType
+from typing import Any, Optional, Tuple
 
+import attr
 from multidict import CIMultiDict, CIMultiDictProxy, MultiDict, MultiDictProxy
 from yarl import URL
 
-from . import hdrs, helpers, http, payload
+from . import hdrs, helpers, http, multipart, payload
 from .client_exceptions import (ClientConnectionError, ClientOSError,
                                 ClientResponseError, ContentTypeError,
-                                InvalidURL)
+                                InvalidURL, ServerFingerprintMismatch)
 from .formdata import FormData
-from .helpers import PY_35, HeadersMixin, SimpleCookie, TimerNoop, noop
-from .http import SERVER_SOFTWARE, HttpVersion10, HttpVersion11, PayloadWriter
+from .helpers import PY_36, HeadersMixin, TimerNoop, noop, reify, set_result
+from .http import SERVER_SOFTWARE, HttpVersion10, HttpVersion11, StreamWriter
 from .log import client_logger
-from .streams import FlowControlStreamReader
+from .streams import StreamReader  # noqa
+from .typedefs import DEFAULT_JSON_DECODER, JSONDecoder, RawHeaders
 
 
+try:
+    import ssl
+except ImportError:  # pragma: no cover
+    ssl = None  # type: ignore
+
 try:
     import cchardet as chardet
 except ImportError:  # pragma: no cover
     import chardet
 
 
-__all__ = ('ClientRequest', 'ClientResponse', 'RequestInfo')
+__all__ = ('ClientRequest', 'ClientResponse', 'RequestInfo', 'Fingerprint')
 
 
-RequestInfo = collections.namedtuple(
-    'RequestInfo', ('url', 'method', 'headers'))
+json_re = re.compile(r'^application/(?:[\w.+-]+?\+)?json')
 
 
-HASHFUNC_BY_DIGESTLEN = {
-    16: md5,
-    20: sha1,
-    32: sha256,
-}
+@attr.s(frozen=True, slots=True)
+class ContentDisposition:
+    type = attr.ib(type=str)
+    parameters = attr.ib(type=MappingProxyType)
+    filename = attr.ib(type=str)
 
 
-_SSL_OP_NO_COMPRESSION = getattr(ssl, "OP_NO_COMPRESSION", 0)
+@attr.s(frozen=True, slots=True)
+class RequestInfo:
+    url = attr.ib(type=URL)
+    method = attr.ib(type=str)
+    headers = attr.ib(type=CIMultiDictProxy)
+    real_url = attr.ib(type=URL)
+
+    @real_url.default
+    def real_url_default(self):
+        return self.url
 
 
-ConnectionKey = namedtuple('ConnectionKey', ['host', 'port', 'ssl'])
+class Fingerprint:
+    HASHFUNC_BY_DIGESTLEN = {
+        16: md5,
+        20: sha1,
+        32: sha256,
+    }
+
+    def __init__(self, fingerprint):
+        digestlen = len(fingerprint)
+        hashfunc = self.HASHFUNC_BY_DIGESTLEN.get(digestlen)
+        if not hashfunc:
+            raise ValueError('fingerprint has invalid length')
+        elif hashfunc is md5 or hashfunc is sha1:
+            raise ValueError('md5 and sha1 are insecure and '
+                             'not supported. Use sha256.')
+        self._hashfunc = hashfunc
+        self._fingerprint = fingerprint
+
+    @property
+    def fingerprint(self):
+        return self._fingerprint
+
+    def check(self, transport):
+        if not transport.get_extra_info('sslcontext'):
+            return
+        sslobj = transport.get_extra_info('ssl_object')
+        cert = sslobj.getpeercert(binary_form=True)
+        got = self._hashfunc(cert).digest()
+        if got != self._fingerprint:
+            host, port, *_ = transport.get_extra_info('peername')
+            raise ServerFingerprintMismatch(self._fingerprint,
+                                            got, host, port)
+
+
+if ssl is not None:
+    SSL_ALLOWED_TYPES = (ssl.SSLContext, bool, Fingerprint, type(None))
+else:  # pragma: no cover
+    SSL_ALLOWED_TYPES = type(None)
+
+
+def _merge_ssl_params(ssl, verify_ssl, ssl_context, fingerprint):
+    if verify_ssl is not None and not verify_ssl:
+        warnings.warn("verify_ssl is deprecated, use ssl=False instead",
+                      DeprecationWarning,
+                      stacklevel=3)
+        if ssl is not None:
+            raise ValueError("verify_ssl, ssl_context, fingerprint and ssl "
+                             "parameters are mutually exclusive")
+        else:
+            ssl = False
+    if ssl_context is not None:
+        warnings.warn("ssl_context is deprecated, use ssl=context instead",
+                      DeprecationWarning,
+                      stacklevel=3)
+        if ssl is not None:
+            raise ValueError("verify_ssl, ssl_context, fingerprint and ssl "
+                             "parameters are mutually exclusive")
+        else:
+            ssl = ssl_context
+    if fingerprint is not None:
+        warnings.warn("fingerprint is deprecated, "
+                      "use ssl=Fingerprint(fingerprint) instead",
+                      DeprecationWarning,
+                      stacklevel=3)
+        if ssl is not None:
+            raise ValueError("verify_ssl, ssl_context, fingerprint and ssl "
+                             "parameters are mutually exclusive")
+        else:
+            ssl = Fingerprint(fingerprint)
+    if not isinstance(ssl, SSL_ALLOWED_TYPES):
+        raise TypeError("ssl should be SSLContext, bool, Fingerprint or None, "
+                        "got {!r} instead.".format(ssl))
+    return ssl
+
+
+@attr.s(slots=True, frozen=True)
+class ConnectionKey:
+    # the key should contain an information about used proxy / TLS
+    # to prevent reusing wrong connections from a pool
+    host = attr.ib(type=str)
+    port = attr.ib(type=int)
+    is_ssl = attr.ib(type=bool)
+    ssl = attr.ib()  # SSLContext or None
+    proxy = attr.ib()  # URL or None
+    proxy_auth = attr.ib()  # BasicAuth
+    proxy_headers_hash = attr.ib(type=int)  # hash(CIMultiDict)
+
+
+def _is_expected_content_type(response_content_type, expected_content_type):
+    if expected_content_type == 'application/json':
+        return json_re.match(response_content_type)
+    return expected_content_type in response_content_type
 
 
 class ClientRequest:
@@ -85,14 +191,10 @@ class ClientRequest:
                  chunked=None, expect100=False,
                  loop=None, response_class=None,
                  proxy=None, proxy_auth=None,
-                 timer=None, session=None, auto_decompress=True,
-                 verify_ssl=None, fingerprint=None, ssl_context=None,
-                 proxy_headers=None):
-
-        if verify_ssl is False and ssl_context is not None:
-            raise ValueError(
-                "Either disable ssl certificate validation by "
-                "verify_ssl=False or specify ssl_context, not both.")
+                 timer=None, session=None,
+                 ssl=None,
+                 proxy_headers=None,
+                 traces=None):
 
         if loop is None:
             loop = asyncio.get_event_loop()
@@ -105,8 +207,8 @@ class ClientRequest:
             url2 = url.with_query(params)
             q.extend(url2.query)
             url = url.with_query(q)
-        self.url = url.with_fragment(None)
         self.original_url = url
+        self.url = url.with_fragment(None)
         self.method = method.upper()
         self.chunked = chunked
         self.compress = compress
@@ -114,9 +216,7 @@ class ClientRequest:
         self.length = None
         self.response_class = response_class or ClientResponse
         self._timer = timer if timer is not None else TimerNoop()
-        self._auto_decompress = auto_decompress
-        self._verify_ssl = verify_ssl
-        self._ssl_context = ssl_context
+        self._ssl = ssl
 
         if loop.get_debug():
             self._source_traceback = traceback.extract_stack(sys._getframe(1))
@@ -129,16 +229,32 @@ class ClientRequest:
         self.update_content_encoding(data)
         self.update_auth(auth)
         self.update_proxy(proxy, proxy_auth, proxy_headers)
-        self.update_fingerprint(fingerprint)
 
         self.update_body_from_data(data)
         if data or self.method not in self.GET_METHODS:
             self.update_transfer_encoding()
         self.update_expect_continue(expect100)
+        if traces is None:
+            traces = []
+        self._traces = traces
+
+    def is_ssl(self):
+        return self.url.scheme in ('https', 'wss')
+
+    @property
+    def ssl(self):
+        return self._ssl
 
     @property
     def connection_key(self):
-        return ConnectionKey(self.host, self.port, self.ssl)
+        proxy_headers = self.proxy_headers
+        if proxy_headers:
+            h = hash(tuple((k, v) for k, v in proxy_headers.items()))
+        else:
+            h = None
+        return ConnectionKey(self.host, self.port, self.is_ssl(),
+                             self.ssl,
+                             self.proxy, self.proxy_auth, h)
 
     @property
     def host(self):
@@ -150,7 +266,8 @@ class ClientRequest:
 
     @property
     def request_info(self):
-        return RequestInfo(self.url, self.method, self.headers)
+        return RequestInfo(self.url, self.method,
+                           self.headers, self.original_url)
 
     def update_host(self, url):
         """Update destination host, port and connection type (ssl)."""
@@ -163,11 +280,6 @@ class ClientRequest:
         if username:
             self.auth = helpers.BasicAuth(username, password or '')
 
-        # Record entire netloc for usage in host header
-
-        scheme = url.scheme
-        self.ssl = scheme in ('https', 'wss')
-
     def update_version(self, version):
         """Convert request version to two elements tuple.
 
@@ -327,7 +439,7 @@ class ClientRequest:
             expect = True
 
         if expect:
-            self._continue = helpers.create_future(self.loop)
+            self._continue = self.loop.create_future()
 
     def update_proxy(self, proxy, proxy_auth, proxy_headers):
         if proxy and not proxy.scheme == 'http':
@@ -338,36 +450,6 @@ class ClientRequest:
         self.proxy_auth = proxy_auth
         self.proxy_headers = proxy_headers
 
-    def update_fingerprint(self, fingerprint):
-        if fingerprint:
-            digestlen = len(fingerprint)
-            hashfunc = HASHFUNC_BY_DIGESTLEN.get(digestlen)
-            if not hashfunc:
-                raise ValueError('fingerprint has invalid length')
-            elif hashfunc is md5 or hashfunc is sha1:
-                warnings.warn('md5 and sha1 are insecure and deprecated. '
-                              'Use sha256.',
-                              DeprecationWarning, stacklevel=2)
-                client_logger.warn('md5 and sha1 are insecure and deprecated. '
-                                   'Use sha256.')
-            self._hashfunc = hashfunc
-        self._fingerprint = fingerprint
-
-    @property
-    def verify_ssl(self):
-        """Do check for ssl certifications?"""
-        return self._verify_ssl
-
-    @property
-    def fingerprint(self):
-        """Expected ssl certificate fingerprint."""
-        return self._fingerprint
-
-    @property
-    def ssl_context(self):
-        """SSLContext instance for https requests."""
-        return self._ssl_context
-
     def keep_alive(self):
         if self.version < HttpVersion10:
             # keep alive not supported at all
@@ -382,25 +464,24 @@ class ClientRequest:
 
         return True
 
-    @asyncio.coroutine
-    def write_bytes(self, writer, conn):
+    async def write_bytes(self, writer, conn):
         """Support coroutines that yields bytes objects."""
         # 100 response
         if self._continue is not None:
-            yield from writer.drain()
-            yield from self._continue
+            await writer.drain()
+            await self._continue
 
         try:
             if isinstance(self.body, payload.Payload):
-                yield from self.body.write(writer)
+                await self.body.write(writer)
             else:
                 if isinstance(self.body, (bytes, bytearray)):
                     self.body = (self.body,)
 
                 for chunk in self.body:
-                    writer.write(chunk)
+                    await writer.write(chunk)
 
-            yield from writer.write_eof()
+            await writer.write_eof()
         except OSError as exc:
             new_exc = ClientOSError(
                 exc.errno,
@@ -416,21 +497,24 @@ class ClientRequest:
         finally:
             self._writer = None
 
-    def send(self, conn):
+    async def send(self, conn):
         # Specify request target:
         # - CONNECT request must send authority form URI
         # - not CONNECT proxy must send absolute form URI
         # - most common is origin form URI
         if self.method == hdrs.METH_CONNECT:
             path = '{}:{}'.format(self.url.raw_host, self.url.port)
-        elif self.proxy and not self.ssl:
+        elif self.proxy and not self.is_ssl():
             path = str(self.url)
         else:
             path = self.url.raw_path
             if self.url.raw_query_string:
                 path += '?' + self.url.raw_query_string
 
-        writer = PayloadWriter(conn.writer, self.loop)
+        writer = StreamWriter(
+            conn.protocol, self.loop,
+            on_chunk_sent=self._on_chunk_request_sent
+        )
 
         if self.compress:
             writer.enable_compression(self.compress)
@@ -458,28 +542,26 @@ class ClientRequest:
             self.headers[hdrs.CONNECTION] = connection
 
         # status + headers
-        status_line = '{0} {1} HTTP/{2[0]}.{2[1]}\r\n'.format(
+        status_line = '{0} {1} HTTP/{2[0]}.{2[1]}'.format(
             self.method, path, self.version)
-        writer.write_headers(status_line, self.headers)
+        await writer.write_headers(status_line, self.headers)
 
-        self._writer = helpers.ensure_future(
-            self.write_bytes(writer, conn), loop=self.loop)
+        self._writer = self.loop.create_task(self.write_bytes(writer, conn))
 
         self.response = self.response_class(
             self.method, self.original_url,
             writer=self._writer, continue100=self._continue, timer=self._timer,
             request_info=self.request_info,
-            auto_decompress=self._auto_decompress
+            traces=self._traces,
+            loop=self.loop,
+            session=self._session
         )
-
-        self.response._post_init(self.loop, self._session)
         return self.response
 
-    @asyncio.coroutine
-    def close(self):
+    async def close(self):
         if self._writer is not None:
             try:
-                yield from self._writer
+                await self._writer
             finally:
                 self._writer = None
 
@@ -489,78 +571,95 @@ class ClientRequest:
                 self._writer.cancel()
             self._writer = None
 
+    async def _on_chunk_request_sent(self, chunk):
+        for trace in self._traces:
+            await trace.send_request_chunk_sent(chunk)
+
 
 class ClientResponse(HeadersMixin):
 
     # from the Status-Line of the response
     version = None  # HTTP-Version
-    status = None   # Status-Code
+    status = None   # type: int  # Status-Code
     reason = None   # Reason-Phrase
 
-    content = None  # Payload stream
-    headers = None  # Response headers, CIMultiDictProxy
-    raw_headers = None  # Response raw headers, a sequence of pairs
+    content = None  # type: StreamReader  # Payload stream
+    _headers = None  # type: CIMultiDictProxy  # Response headers
+    _raw_headers = None  # type: RawHeaders  # Response raw headers
 
     _connection = None  # current connection
-    flow_control_class = FlowControlStreamReader  # reader flow control
-    _reader = None     # input stream
     _source_traceback = None
     # setted up by ClientRequest after ClientResponse object creation
     # post-init stage allows to not change ctor signature
-    _loop = None
     _closed = True  # to allow __del__ for non-initialized properly response
-    _session = None
+    _released = False
 
     def __init__(self, method, url, *,
-                 writer=None, continue100=None, timer=None,
-                 request_info=None, auto_decompress=True):
+                 writer, continue100, timer,
+                 request_info,
+                 traces, loop, session):
         assert isinstance(url, URL)
 
         self.method = method
-        self.headers = None
         self.cookies = SimpleCookie()
 
-        self._url = url
-        self._content = None
+        self._real_url = url
+        self._url = url.with_fragment(None)
+        self._body = None  # type: bytes
         self._writer = writer
-        self._continue = continue100
+        self._continue = continue100  # None by default
         self._closed = True
         self._history = ()
         self._request_info = request_info
         self._timer = timer if timer is not None else TimerNoop()
-        self._auto_decompress = auto_decompress
-
-    @property
-    def url(self):
-        return self._url
-
-    @property
-    def url_obj(self):
-        warnings.warn(
-            "Deprecated, use .url #1654", DeprecationWarning, stacklevel=2)
-        return self._url
-
-    @property
-    def host(self):
-        return self._url.host
-
-    @property
-    def _headers(self):
-        return self.headers
-
-    @property
-    def request_info(self):
-        return self._request_info
-
-    def _post_init(self, loop, session):
+        self._cache = {}  # required for @reify method decorator
+        self._traces = traces
         self._loop = loop
         self._session = session  # store a reference to session #1985
         if loop.get_debug():
             self._source_traceback = traceback.extract_stack(sys._getframe(1))
 
+    @reify
+    def url(self) -> URL:
+        return self._url
+
+    @reify
+    def url_obj(self) -> URL:
+        warnings.warn(
+            "Deprecated, use .url #1654", DeprecationWarning, stacklevel=2)
+        return self._url
+
+    @reify
+    def real_url(self) -> URL:
+        return self._real_url
+
+    @reify
+    def host(self) -> str:
+        return self._url.host
+
+    @reify
+    def headers(self) -> CIMultiDictProxy:
+        return self._headers
+
+    @reify
+    def raw_headers(self) -> RawHeaders:
+        return self._raw_headers
+
+    @reify
+    def request_info(self) -> RequestInfo:
+        return self._request_info
+
+    @reify
+    def content_disposition(self) -> Optional[ContentDisposition]:
+        raw = self._headers.get(hdrs.CONTENT_DISPOSITION)
+        if raw is None:
+            return None
+        disposition_type, params = multipart.parse_content_disposition(raw)
+        params = MappingProxyType(params)
+        filename = multipart.content_disposition_filename(params)
+        return ContentDisposition(disposition_type, params, filename)
+
     def __del__(self, _warnings=warnings):
-        if self._loop is None:
-            return  # not started
         if self._closed:
             return
 
@@ -568,16 +667,19 @@ class ClientResponse(HeadersMixin):
             self._connection.release()
             self._cleanup_writer()
 
-            # warn
-            if __debug__:
-                if self._loop.get_debug():
-                    _warnings.warn("Unclosed response {!r}".format(self),
-                                   ResourceWarning)
-                    context = {'client_response': self,
-                               'message': 'Unclosed response'}
-                    if self._source_traceback:
-                        context['source_traceback'] = self._source_traceback
-                    self._loop.call_exception_handler(context)
+            if self._loop.get_debug():
+                if PY_36:
+                    kwargs = {'source': self}
+                else:
+                    kwargs = {}
+                _warnings.warn("Unclosed response {!r}".format(self),
+                               ResourceWarning,
+                               **kwargs)
+                context = {'client_response': self,
+                           'message': 'Unclosed response'}
+                if self._source_traceback:
+                    context['source_traceback'] = self._source_traceback
+                self._loop.call_exception_handler(context)
 
     def __repr__(self):
         out = io.StringIO()
@@ -598,42 +700,73 @@ class ClientResponse(HeadersMixin):
     def connection(self):
         return self._connection
 
-    @property
-    def history(self):
+    @reify
+    def history(self) -> Tuple['ClientResponse', ...]:
         """A sequence of of responses, if redirects occurred."""
         return self._history
 
-    @asyncio.coroutine
-    def start(self, connection, read_until_eof=False):
+    @reify
+    def links(self) -> MultiDictProxy:
+        links_str = ", ".join(self.headers.getall("link", []))
+
+        links = MultiDict()  # type: MultiDict
+
+        if not links_str:
+            return MultiDictProxy(links)
+
+        for val in re.split(r",(?=\s*<)", links_str):
+            match = re.match(r"\s*<(.*)>(.*)", val)
+            if match is None:  # pragma: no cover
+                # the check exists to suppress mypy error
+                continue
+            url, params_str = match.groups()
+            params = params_str.split(";")[1:]
+
+            link = MultiDict()  # type: MultiDict
+
+            for param in params:
+                match = re.match(
+                    r"^\s*(\S*)\s*=\s*(['\"]?)(.*?)(\2)\s*$",
+                    param, re.M
+                )
+                if match is None:  # pragma: no cover
+                    # the check exists to suppress mypy error
+                    continue
+                key, _, value, _ = match.groups()
+
+                link.add(key, value)
+
+            key = link.get("rel", url)
+
+            link.add("url", self.url.join(URL(url)))
+
+            links.add(key, MultiDictProxy(link))
+
+        return MultiDictProxy(links)
+
+    async def start(self, connection):
         """Start response processing."""
         self._closed = False
         self._protocol = connection.protocol
         self._connection = connection
 
-        connection.protocol.set_response_params(
-            timer=self._timer,
-            skip_payload=self.method.lower() == 'head',
-            skip_status_codes=(204, 304),
-            read_until_eof=read_until_eof,
-            auto_decompress=self._auto_decompress)
-
         with self._timer:
             while True:
                 # read response
                 try:
-                    (message, payload) = yield from self._protocol.read()
+                    message, payload = await self._protocol.read()
                 except http.HttpProcessingError as exc:
                     raise ClientResponseError(
                         self.request_info, self.history,
-                        code=exc.code,
+                        status=exc.code,
                         message=exc.message, headers=exc.headers) from exc
 
                 if (message.code < 100 or
                         message.code > 199 or message.code == 101):
                     break
 
-                if self._continue is not None and not self._continue.done():
-                    self._continue.set_result(True)
+                if self._continue is not None:
+                    set_result(self._continue, True)
                     self._continue = None
 
         # payload eof handler
@@ -645,8 +778,8 @@ class ClientResponse(HeadersMixin):
         self.reason = message.reason
 
         # headers
-        self.headers = CIMultiDictProxy(message.headers)
-        self.raw_headers = tuple(message.raw_headers)
+        self._headers = message.headers  # type is CIMultiDictProxy
+        self._raw_headers = message.raw_headers  # type is Tuple[bytes, bytes]
 
         # payload
         self.content = payload
@@ -678,10 +811,12 @@ class ClientResponse(HeadersMixin):
         self._cleanup_writer()
 
     @property
-    def closed(self):
+    def closed(self) -> bool:
         return self._closed
 
-    def close(self):
+    def close(self) -> None:
+        if not self._released:
+            self._notify_content()
         if self._closed:
             return
 
@@ -693,9 +828,10 @@ class ClientResponse(HeadersMixin):
             self._connection.close()
             self._connection = None
         self._cleanup_writer()
-        self._notify_content()
 
-    def release(self):
+    def release(self) -> Any:
+        if not self._released:
+            self._notify_content()
         if self._closed:
             return noop()
 
@@ -705,15 +841,14 @@ class ClientResponse(HeadersMixin):
             self._connection = None
 
         self._cleanup_writer()
-        self._notify_content()
         return noop()
 
-    def raise_for_status(self):
+    def raise_for_status(self) -> None:
         if 400 <= self.status:
             raise ClientResponseError(
                 self.request_info,
                 self.history,
-                code=self.status,
+                status=self.status,
                 message=self.reason,
                 headers=self.headers)
 
@@ -725,68 +860,76 @@ class ClientResponse(HeadersMixin):
 
     def _notify_content(self):
         content = self.content
-        if content and content.exception() is None and not content.is_eof():
+        if content and content.exception() is None:
             content.set_exception(
                 ClientConnectionError('Connection closed'))
+        self._released = True
 
-    @asyncio.coroutine
-    def wait_for_close(self):
+    async def wait_for_close(self) -> None:
         if self._writer is not None:
             try:
-                yield from self._writer
+                await self._writer
             finally:
                 self._writer = None
         self.release()
 
-    @asyncio.coroutine
-    def read(self):
+    async def read(self) -> bytes:
         """Read response payload."""
-        if self._content is None:
+        if self._body is None:
             try:
-                self._content = yield from self.content.read()
-            except:
+                self._body = await self.content.read()
+                for trace in self._traces:
+                    await trace.send_response_chunk_received(self._body)
+            except BaseException:
                 self.close()
                 raise
+        elif self._released:
+            raise ClientConnectionError('Connection closed')
 
-        return self._content
+        return self._body
 
-    def _get_encoding(self):
+    def get_encoding(self) -> str:
         ctype = self.headers.get(hdrs.CONTENT_TYPE, '').lower()
-        mtype, stype, _, params = helpers.parse_mimetype(ctype)
+        mimetype = helpers.parse_mimetype(ctype)
 
-        encoding = params.get('charset')
+        encoding = mimetype.parameters.get('charset')
+        if encoding:
+            try:
+                codecs.lookup(encoding)
+            except LookupError:
+                encoding = None
         if not encoding:
-            if mtype == 'application' and stype == 'json':
+            if mimetype.type == 'application' and mimetype.subtype == 'json':
                 # RFC 7159 states that the default encoding is UTF-8.
                 encoding = 'utf-8'
             else:
-                encoding = chardet.detect(self._content)['encoding']
+                encoding = chardet.detect(self._body)['encoding']
         if not encoding:
             encoding = 'utf-8'
 
         return encoding
 
-    @asyncio.coroutine
-    def text(self, encoding=None, errors='strict'):
+    async def text(self,
+                   encoding: Optional[str]=None, errors: str='strict') -> str:
         """Read response payload and decode."""
-        if self._content is None:
-            yield from self.read()
+        if self._body is None:
+            await self.read()
 
         if encoding is None:
-            encoding = self._get_encoding()
+            encoding = self.get_encoding()
 
-        return self._content.decode(encoding, errors=errors)
+        return self._body.decode(encoding, errors=errors)
 
-    @asyncio.coroutine
-    def json(self, *, encoding=None, loads=json.loads,
-             content_type='application/json'):
+    async def json(self, *, encoding: str=None,
+                   loads: JSONDecoder=DEFAULT_JSON_DECODER,
+                   content_type: Optional[str]='application/json') -> Any:
         """Read and decodes JSON response."""
-        if self._content is None:
-            yield from self.read()
+        if self._body is None:
+            await self.read()
 
         if content_type:
             ctype = self.headers.get(hdrs.CONTENT_TYPE, '').lower()
-            if content_type not in ctype:
+            if not _is_expected_content_type(ctype, content_type):
                 raise ContentTypeError(
                     self.request_info,
                     self.history,
@@ -794,23 +937,20 @@ class ClientResponse(HeadersMixin):
                              'unexpected mimetype: %s' % ctype),
                     headers=self.headers)
 
-        stripped = self._content.strip()
+        stripped = self._body.strip()
         if not stripped:
             return None
 
         if encoding is None:
-            encoding = self._get_encoding()
+            encoding = self.get_encoding()
 
         return loads(stripped.decode(encoding))
 
-    if PY_35:
-        @asyncio.coroutine
-        def __aenter__(self):
-            return self
+    async def __aenter__(self) -> 'ClientResponse':
+        return self
 
-        @asyncio.coroutine
-        def __aexit__(self, exc_type, exc_val, exc_tb):
-            # similar to _RequestContextManager, we do not need to check
-            # for exceptions, response object can closes connection
-            # is state is broken
-            self.release()
+    async def __aexit__(self, exc_type, exc_val, exc_tb):
+        # similar to _RequestContextManager, we do not need to check
+        # for exceptions, response object can closes connection
+        # is state is broken
+        self.release()
